/**
 * Created by John on 7/14/2017.
 *
 * This class is only used for TESTING!!!!!
 *
 */
var rp = require("request-promise");
var _ = require("lodash");
var moment = require("moment");
var fs = require("fs");
var sleep = require("thread-sleep"); //perf testing only!!!
var API = require("../services/API");
var ASN = require("../services/ASN");


function AsnHelper() {
    this.config      = require("../config/server.qa.js");
    //default items for test-rabbit.js
    this.items = [
        {  vendorMaterial: '42307', description: 'TAZICEF 1GM VL 25X20ML', sapMaterial: '50022316'},
        {  vendorMaterial: '42308', description: 'GENTAMICIN 40MG/ML FTV 25X2ML', sapMaterial: '50022262'},
        {  vendorMaterial: '42309', description: 'VANCOMYCIN 500MG MDV 10/BX', sapMaterial: '50012561'},
        {  vendorMaterial: '42310', description: 'VANCOMYCIN 1GM SDV 10/PK', sapMaterial: '50012550'},
        {  vendorMaterial: '42342', description: 'CEFAZOLIN 1GM VL 25X10ML', sapMaterial: '50012540'}
    ];
}
/**
 * Generates an ASN JSON object for each item in the objects "items" array
 *
 * @param {Object} clinic object who will own the ASN.  If null, defaults to .9999BM
 * @returns {Array} Collection od ASN JSON
 */
AsnHelper.prototype.buildAsns = function(clinic) {
    var self = this;
    var lines = [];
    var idx = 0;
    var alphabet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'.split('');

    var asnId = ((clinic) ? clinic.clinicId.replace(/\D/g,'') : "") + (new moment()).format("HHmm");

    var po = _.sampleSize(alphabet, 4).join("") + "-" + ((clinic) ? clinic.clinicId.replace(/\D/g,'') :   (new moment()).format("mmss"));

    _.forEach(self.items, function (item) {
        var line = {
            asnId: asnId,
            asnDate: new moment(),
            poNumber: po,
            poDate: null,
            vendorId: "4432",
            vendorName: "AB",
            clinicId: (clinic) ? clinic.clinicId : ".9999BM",
            shipToAccount: "09777",
            shipToName: "Tech 1",
            shipFromId: "7777",
            shipVia: "AIR",
            vendorItemId: item.vendorMaterial,
            ndc: null,
            sapMaterialId: null,
            serializationNumber: null,
            shipQty: 100,
            unitPrice: 5.99,
            uom: "EA",
            description: item.description,
            lot: ((idx % 2 === 0) ? "A123" : null),
            expiration: ((idx % 2 === 0) ? (new moment().add(100, 'days')) : null),
            invoiceNumber: "9999",
            lineNumber: "" + (1 + idx),
            trackingId: "1Z37319691464" + idx++
        };
        lines.push(line);
    });
    return lines;
};
/**
 * Generates a collection of Receipt objects based on this.items
 * @param {String} clinic Clinic ID for the record
 * @returns {Array} Collections of Receipt objects
 */
AsnHelper.prototype.buildReceipt = function(clinic) {
    var self = this;
    var lines = [];
    var idx = 0;
    _.forEach(self.items, function (item) {
        var line = {
            "consumerId":"",
            "poNumber": item.poNumber.split("-")[1],
            "poPrefix":item.poNumber.split("-")[0],
            "type":"RECEIPT",
            "userId": (clinic) ? clinic.replace("\.", "%") : "%9999BM",
            "catType":"",
            "clinicId": (clinic) ? clinic : ".9999BM",
            "vendorMaterial": item.vendorMaterial,
            "timestamp": (new moment().add(idx, 'minutes')),
            "countType":"",
            "description": item.description,
            "deviceId": ((clinic) ? "152615225" + clinic.replace(/\D/g,'') : "15261522502149"),
            "sapMaterial": item.sapMaterial,
            "id":0,
            "devicePk":idx++,
            "count": Math.floor(Math.random() * 100),
            "cntInd":false,
            "asnId": item.asnId
        };
        lines.push(line);
    });
    return lines;
};
/**
 * Sends ASN data into the queues via the WEB API
 *
 * @param {String} user
 * @param {String} pass
 * @param {Function} cb callback(err, response) after API POST
 * @param {Object} clinic (optional) object with clinic attributes, especially clinicId
 * @param {Array} asns (optional) list to publish.  If missing, will generate defaults.
 * @param {String} endpoint (optional) URL to which to post the ASN
 */
AsnHelper.prototype.publishViaApi = function(user, pass, cb, clinic, asns, endpoint) {
    var self = this;
    var lines = asns;
    if (!lines) lines = self.buildAsns(clinic);
    //send lines
    var port = process.env.PORT || 8080;
    var url = "http://localhost:" + port;
    if (endpoint) url = endpoint;
    var tkn = null;
    var tknOptions = {
        uri: (url.includes("/api") ? url.substring(0, url.indexOf("/api")) : url) + "/login",
        method: 'POST',
        headers: {
            'User-Agent': 'fmc (node/' + process.version + ')',
            'Content-Type': 'application/json'
        },
        auth: {
            'user': user,
            'pass': pass
        },
        json: true,
        resolveWithFullResponse: true
    };
    var options = {
        uri: (url.includes("/api") ? url : url + "/api/v1/asn"),
        method: 'POST',
        headers: {
            'User-Agent': 'fmc (node/' + process.version + ')',
            'Content-Type': 'application/json'
        },
        auth: {
            bearer: tkn
        },
        body: lines,
        json: true,
        resolveWithFullResponse: true
    };
    rp(tknOptions).then(function (tknResponse) {
        options.auth.bearer = tknResponse.body.token;
        rp(options)
            .then(function (response) {
                if (cb) cb(null, response);
            }).catch(function (err) {
            if (cb) cb(err, null);
            else throw err;
        });
    }).catch(function (tknerr) {
        if (cb) cb(tknerr, null);
        else throw tknerr;
    });

};

/**
 * Sends receipt data into the db via the WEB API
 *
 * @param {Array} receipts staged data
 * @param {String} endpoint API to send
 * @param {Function} cb callback(err, response) after API post
 */
AsnHelper.prototype.receiveViaApi = function(receipts, endpoint, cb){
    var self = this;
    var service = new API(self.config);
    service.saveReceipts(receipts, endpoint).then(function (response) {
        if (cb) cb(null, response);
    }).catch(function (err) {
        if (cb) cb(err, null);
        else throw err;
    });

};



/**
 * Main driver for creating dynamic test ASN data
 *
 * @param {String} endpoint
 * @param {Number} clinics number of clinics to sample
 * @param {Number} lines number of line items per ASN
 * @param {String} filename Full path to temp file to stage the data
 * @param {Function} cb Callback for success/failure
 */
AsnHelper.prototype.generateAsns = function (endpoint, clinics, lines, filename, cb) {
        var self = this;
        var service = new API(self.config);
        service.getClinics(endpoint).then(function(clinicsResponse){
            var allClinics = clinicsResponse.body;
            var sampleClinics = _.sampleSize(allClinics, clinics);
            service.getItems(endpoint).then(function(itemResponse){
                var allItems = itemResponse.body;
                self.items = _.sampleSize(allItems, lines);
                for (var i = 0; i < sampleClinics.length; i++) {
                    var asns = self.buildAsns(sampleClinics[i]);
                    fs.appendFileSync(filename, ((i===0) ? "" : ",\n") + JSON.stringify(asns));
                }
                cb(null, filename);
            }).catch(function(iErr){
                console.error(JSON.stringify(iErr));
                cb(iErr, null);
            });

        }).catch(function(cErr){
            console.error(JSON.stringify(cErr));
            cb(cErr, null);
        });
};
/**
 * Main driver for creating dynamic test Receipt data
 *
 * @param {String} endpoint
 * @param {Number} clinics number of clinics to sample
 * @param {Number} lines number of line items per ASN
 * @param {String} filename Full path to temp file to stage the data
 * @param {Function} cb Callback for success/failure
 */
AsnHelper.prototype.generateReceipts = function (endpoint, clinics, lines, filename, cb) {
    var self = this;
    var service = new ASN(self.config);
    service.getAsns().then(function(asnResponse){
        var allAsns = asnResponse.body;
        var allClinics = _.map(allAsns, "clinicId");
        var sampleClinics = _.sampleSize(allClinics, clinics);
        for (var i = 0; i < sampleClinics.length; i++) {
            var clinicId = sampleClinics[i];
            var asns = _.filter(allAsns, { clinicId: clinicId });
            var sampleAsns = _.sampleSize(asns, lines);
            self.items = [];
            for(var a = 0; a < sampleAsns.length; a++) {
                var asn = sampleAsns[a];
                self.items.push({asnId: asn.id, vendorMaterial: asn.vendorItemId, sapMaterial: asn.sapMaterialId, description: asn.description, poNumber: asn.poNumber });
            }
            var receipts = self.buildReceipt(clinicId);
            fs.appendFileSync(filename, ((i===0) ? "" : ",\n") + JSON.stringify(receipts));
        }
        cb(null, filename);
    }).catch(function(iErr){
        console.error(JSON.stringify(iErr));
        cb(iErr, null);
    });


};

/**
 * Batches the test data and POSTs to the resulting endpoint
 * @param {String} realFileName Full path containing the staged data
 * @param {String} user Username for token authentication
 * @param {String} password Password for token authentication
 * @param {Number} batch umber of records to post at once
 * @param {Number} delay milliseconds to wait between POSTs
 * @param {String} endpoint
 */
AsnHelper.prototype.sendAsns = function(realFileName, user, password, batch, delay, endpoint) {
    var self = this;
    var data = this.getBatches(realFileName, batch);
    //data is array of arrays of the requested batch size
    for (var d = 0; d < data.length; d++) {
        if (Number(delay) > 0) {
            sleep(Number(delay));
        }
        self.publishViaApi(user, password, function (err, response){
            if (err) console.error(JSON.stringify(err))
            else console.log("Published " + data[d].length + " ASNs ");
        }, null, data[d], endpoint );
    }
};
/**
 * Batches the test data and POSTs to the resulting endpoint
 * @param {String} realFileName Full path containing the staged data
 * @param {Number} batch umber of records to post at once
 * @param {Number} delay milliseconds to wait between POSTs
 * @param {String} endpoint
 */
AsnHelper.prototype.sendReceipts = function( realFileName, batch, delay, endpoint ) {
    var self = this;
    var data = this.getBatches(realFileName, batch);
    //data is array of arrays of the requested batch size
    //data is array of arrays of the requested batch size
    for (var d = 0; d < data.length; d++) {
        if (Number(delay) > 0) {
            sleep(Number(delay));
        }
        self.receiveViaApi(data[d], endpoint, function (err, response){
            if (err) console.error(err)
            else console.log("Published Receipts ");
        })
    }
};
/**
 * Opens the data file and converts into an array of batches
 * @param {String} filename full path to data stage file
 * @param {Number} batchSize Number of records to post at once
 * @returns {Array} Collection of batches to post
 */
AsnHelper.prototype.getBatches = function(filename, batchSize) {
    var self = this;
    var data = fs.readFileSync(filename);
    if (data) data = JSON.parse(data);
    //data is array of arrays
    var batches = [];
    while (data.length > 0) {
        batches.push(self.buildBatch(data, batchSize));
    }
    return batches;
};
/**
 * Breaks up arrays of data into new arrays of data based on POST batch size
 * @param {Array} collection of data generated
 * @param {Number} batchSize size of "chunks" to be posted
 * @returns {Array} Collection of data to be submitted at once
 */
AsnHelper.prototype.buildBatch = function(collection, batchSize) {
    var line = collection.pop();
    if (line.length > batchSize) {
        //split this line.  Add the remainder back, returning the batch size
        var out = _.slice(line, 0, batchSize);
        line = _.drop(line, batchSize);
        collection.push(line);
        return out;
    } else if (line.length === batchSize) {
        return line;
    } else {
        //multiple lines make up a batch size
        var combined = line;
        while (combined.length < batchSize && collection.length > 0 ) {
            combined = _.concat(combined, collection.pop());
        }
        //each pop is adding multiple records at a time,
        // so even with the "while" condition, we could over step batch size
        if (combined.length > batchSize) {
            //split this line.  Add the remainder back, returning the batch size
            var out = _.slice(combined, 0, batchSize);
            combined = _.drop(combined, batchSize);
            collection.push(combined);
            return out;
        } else  {
            return combined;
        }
    }
};



module.exports = AsnHelper;