var _ = require("lodash");
var SAP = require("../services/SAP");
var API = require("../services/API");

function ReceiptHelper(config) {
    this.config = config;
}

/**
 * Retrieves pending Receipts from the API service
 */
ReceiptHelper.prototype.getReceipts = function() {
    var self = this;
    var service = new API(self.config);
    return service.getReceipts();
};

/**
 * Updates Receipt flags via the API service
 *
 * @param ids Database keys to update
 */
ReceiptHelper.prototype.updateReceipts = function(ids) {
    var self = this;
    var service = new API(self.config);
    return service.updateReceipts(ids);
};

/**
 * Sends the Receipts to SAP via the SAP service
 *
 * @param records data to send to SAP
 * @returns {Promise}
 */
ReceiptHelper.prototype.toSAP = function(records) {
    var self = this;
    var service = new SAP(self.config);
    return service.toFile(records);
};

/**
 * Pulls records from API
 * Formats for SAP
 * Sends to SAP
 * Updates flag via API
 *
 * @param {Boolean} isTest  If running under unittest (true) or live (false)
 * @returns {Promise}
 */
ReceiptHelper.prototype.orchestrate = function(isTest) {
  var self = this;
  return new Promise(function(resolve, reject){
      self.getReceipts().then(function(response){
          var records = response.body;
          if (records && Array.isArray(records)) {
              var ids = [];
              for (var i=0; i < records.length; i++) {
                  var record = records[i];
                  ids.push(record.id);
                  delete record.id;
              }
              if (ids.length > 0) {
                  self.toSAP(records).then(function (sapResponse) {
                      if (sapResponse.status === 200) {
                          if (isTest) { //do not update records
                              resolve(true);
                          } else {
                              self.updateReceipts(ids).then(function (apiResponse) {
                                  resolve(true);
                              }).catch(function (apiErr) {
                                  reject(apiErr);
                              });
                          }
                      }
                  }).catch(function (sapErr) {
                      reject(sapErr);
                  });
              } else {
                  resolve(true);
              }
          } else {
              resolve(false);
          }
      }).catch(function(err){
          reject(err);
      })
  });

};

module.exports = ReceiptHelper;